@@ -1,22 +1,22 @@
-/* vi:set sw=2 sts=2 ts=2 et ai: */
+/* vi:set et ai sw=2 sts=2 ts=2: */
 /*-
- * Copyright (c) 2008 Jasper Huijsmans <jasper@xfce.org>.
- * Copyright (c) 2008 Jannis Pohlmann <jannis@xfce.org>.
+ * Copyright (c) 2008      Jasper Huijsmans <jasper@xfce.org>
+ * Copyright (c) 2008-2010 Jannis Pohlmann <jannis@xfce.org>
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or (at
- * your option) any later version.
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of
+ * the License, or (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
- * MA  02111-1307  USA
+ * You should have received a copy of the GNU General Public
+ * License along with this program; if not, write to the Free
+ * Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -60,8 +60,7 @@
 enum
 {
   PROP_0,
-  PROP_MENU_FILENAME,
-  PROP_CATEGORY,
+  PROP_MENU_FILENAME
 };
 
 
@@ -78,7 +77,12 @@ static void       xfce_appfinder_window_set_property       (GObject
                                                             const GValue             *value,
                                                             GParamSpec               *pspec);
 static void       _xfce_appfinder_window_closed            (XfceAppfinderWindow      *window);
-static gpointer   _xfce_appfinder_window_reload_menu       (XfceAppfinderWindow      *window);
+static void       _xfce_appfinder_window_load_menu         (XfceAppfinderWindow      *window,
+                                                            GarconMenu               *menu,
+                                                            const gchar              *category,
+                                                            gint                     *counter,
+                                                            gboolean                  is_root,
+                                                            gboolean                  is_category);
 static void       _xfce_appfinder_window_entry_changed     (GtkEditable              *editable,
                                                             XfceAppfinderWindow      *window);
 static void       _xfce_appfinder_window_entry_activated   (GtkEntry                 *entry,
@@ -144,8 +148,6 @@ struct _XfceAppfinderWindow
 
   GarconMenu    *menu;
   gchar         *menu_filename;
-
-  GThread       *reload_thread;
 };
 
 
@@ -180,15 +182,6 @@ xfce_appfinder_window_class_init (XfceAppfinderWindowClass *klass)
                                                         NULL,
                                                         G_PARAM_CONSTRUCT_ONLY |
                                                         G_PARAM_WRITABLE));
-
-  g_object_class_install_property (gobject_class,
-                                   PROP_CATEGORY,
-                                   g_param_spec_string ("category",
-                                                        "category",
-                                                        "category",
-                                                        DEFAULT_CATEGORY,
-                                                        G_PARAM_CONSTRUCT |
-                                                        G_PARAM_READWRITE));
 }
 
 
@@ -214,7 +207,6 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
 
   window->menu = NULL;
   window->menu_filename = NULL;
-  window->reload_thread = NULL;
   window->categories_group = NULL;
   window->current_category = NULL;
 
@@ -351,6 +343,14 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
                         GTK_DIALOG (window)->action_area);
   gtk_box_pack_start (GTK_BOX (hbox2), GTK_DIALOG (window)->action_area, TRUE, TRUE, 0);
   g_object_unref (G_OBJECT (GTK_DIALOG (window)->action_area));
+
+  if (G_UNLIKELY (window->menu_filename != NULL))
+    window->menu = garcon_menu_new_for_path (window->menu_filename);
+  else
+    window->menu = garcon_menu_new_applications ();
+
+  /* reload the menu when necessary */
+  g_signal_connect_swapped (window->menu, "reload-required", G_CALLBACK (xfce_appfinder_window_reload), window);
 }
 
 
@@ -359,8 +359,12 @@ static void
 xfce_appfinder_window_constructed (GObject *object)
 {
   XfceAppfinderWindow *window = XFCE_APPFINDER_WINDOW (object);
+  gchar               *category;
 
-  xfconf_g_property_bind (window->channel, "/category", G_TYPE_STRING, G_OBJECT (window), "category");
+  category = xfconf_channel_get_string (window->channel, "/category", NULL);
+  if (category != NULL)
+    _xfce_appfinder_window_set_category (window, category);
+  g_free (category);
 }
 
 
@@ -397,14 +401,8 @@ xfce_appfinder_window_get_property (GObject    *object,
                                     GValue     *value,
                                     GParamSpec *pspec)
 {
-  XfceAppfinderWindow *window = XFCE_APPFINDER_WINDOW (object);
-
   switch (prop_id)
     {
-    case PROP_CATEGORY:
-      g_value_set_string (value, window->current_category);
-      break;
-
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -428,10 +426,6 @@ xfce_appfinder_window_set_property (GObject      *object,
       window->menu_filename = g_strdup (g_value_get_string (value));
       break;
 
-    case PROP_CATEGORY:
-      _xfce_appfinder_window_set_category (window, g_value_get_string (value));
-      break;
-
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -451,7 +445,45 @@ xfce_appfinder_window_new (const gchar *filename)
 void
 xfce_appfinder_window_reload (XfceAppfinderWindow *window)
 {
-  _xfce_appfinder_window_reload_menu (window);
+  GtkWidget *button;
+  GError    *error = NULL;
+  gint       counter = 0;
+
+  g_return_if_fail (XFCE_IS_APPFINDER_WINDOW (window));
+
+  gtk_list_store_clear (window->list_store);
+
+  if (!garcon_menu_load (window->menu, NULL, &error))
+    {
+      xfce_dialog_show_error (GTK_WINDOW (window), error, _("Could not load menu from %s"), window->menu_filename);
+      g_error_free (error);
+      return;
+    }
+
+  if (GTK_IS_WIDGET (window->categories_box))
+    gtk_widget_destroy (window->categories_box);
+
+  window->categories_box = gtk_vbox_new (FALSE, 6);
+  gtk_container_add (GTK_CONTAINER (window->categories_alignment), window->categories_box);
+  gtk_widget_show (window->categories_box);
+
+  button = gtk_radio_button_new_with_label (NULL, _("All"));
+  gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
+  g_signal_connect_swapped (button, "toggled", G_CALLBACK (_xfce_appfinder_window_category_changed), window);
+  g_signal_connect (button, "key-press-event", G_CALLBACK (_xfce_appfinder_window_radio_key_pressed), window);
+  gtk_container_add (GTK_CONTAINER (window->categories_box), button);
+  gtk_widget_show (button);
+
+  if (window->current_category == NULL || g_utf8_collate (window->current_category, _("All")) == 0)
+    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
+
+  window->categories_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (button));
+
+  _xfce_appfinder_window_load_menu (window, window->menu, NULL, &counter, TRUE, FALSE);
+
+  /* reset the current category when the previous one doesn't exist anymore */
+  if (window->current_category != NULL && gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button)))
+    _xfce_appfinder_window_set_category (window, gtk_button_get_label (GTK_BUTTON (button)));
 }
 
 
@@ -641,6 +673,7 @@ _xfce_appfinder_window_cursor_changed (GtkTreeView         *tree_view,
     {
       gtk_widget_set_sensitive (window->execute_button, TRUE);
       gtk_widget_grab_default (window->execute_button);
+      g_object_unref (G_OBJECT (item));
     }
   else
     {
@@ -678,6 +711,7 @@ _xfce_appfinder_window_drag_data_get (GtkWidget           *widget,
     return;
 
   uri = garcon_menu_item_get_uri (item);
+  g_object_unref (G_OBJECT (item));
   if (G_LIKELY (uri != NULL))
     {
       gtk_selection_data_set (data, data->target, 8, (guchar *) uri, strlen (uri));
@@ -690,42 +724,78 @@ _xfce_appfinder_window_drag_data_get (GtkWidget           *widget,
 static void
 _xfce_appfinder_window_execute (XfceAppfinderWindow *window)
 {
-  GarconMenuItem   *item;
-  GtkTreeSelection *selection;
-  GtkTreeModel     *model;
-  GtkTreeIter       iter;
-  GdkScreen        *screen;
-  GError           *error = NULL;
-  gchar            *command, *uri;
+  GarconMenuItem    *item;
+  GtkTreeSelection  *selection;
+  GtkTreeModel      *model;
+  GtkTreeIter        iter;
+  GError            *error = NULL;
+  const gchar       *command, *p;
+  GString           *string;
+  gchar            **argv;
+  gboolean           result = FALSE;
 
   g_return_if_fail (XFCE_IS_APPFINDER_WINDOW (window));
 
   selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (window->tree_view));
 
-  if (G_UNLIKELY (selection == NULL || !gtk_tree_selection_get_selected (selection, &model, &iter)))
+  if (selection == NULL
+      || !gtk_tree_selection_get_selected (selection, &model, &iter))
     return;
 
   gtk_tree_model_get (model, &iter, ITEM_COLUMN, &item, -1);
-
   if (G_UNLIKELY (item == NULL))
     return;
 
-  uri = garcon_menu_item_get_uri (item);
-  command = g_strconcat ("exo-open ", uri, NULL);
-  g_free (uri);
+  command = garcon_menu_item_get_command (item);
+  if (command == NULL || *command == '\0')
+    return;
 
-  screen = xfce_gdk_screen_get_active (NULL);
+  string = g_string_sized_new (100);
 
-  if (G_UNLIKELY (!xfce_spawn_command_line_on_screen (screen, command, FALSE, TRUE, &error)))
+  if (garcon_menu_item_requires_terminal (item))
+    g_string_append (string, "exo-open --launch TerminalEmulator ");
+
+  /* expand the field codes */
+  for (p = command; *p != '\0'; ++p)
     {
-      xfce_dialog_show_error (GTK_WINDOW (window), error,
-           _("Could not execute application %s."), garcon_menu_element_get_name (GARCON_MENU_ELEMENT (item)));
+      if (G_UNLIKELY (p[0] == '%' && p[1] != '\0'))
+        {
+          switch (*++p)
+            {
+            case '%':
+              g_string_append_c (string, '%');
+              break;
 
-      if (error != NULL)
-        g_error_free (error);
+            /* skip all the other %? values for now we don't have dnd anyways */
+            }
+        }
+      else
+        {
+          g_string_append_c (string, *p);
+        }
     }
 
-  g_free (command);
+  if (g_shell_parse_argv (string->str, NULL, &argv, &error))
+    {
+      result = xfce_spawn_on_screen (gtk_window_get_screen (GTK_WINDOW (window)),
+                                     garcon_menu_item_get_path (item),
+                                     argv, NULL, G_SPAWN_SEARCH_PATH,
+                                     garcon_menu_item_supports_startup_notification (item),
+                                     gtk_get_current_event_time (),
+                                     garcon_menu_item_get_icon_name (item),
+                                     &error);
+
+      g_strfreev (argv);
+    }
+
+  if (G_UNLIKELY (!result))
+    {
+      xfce_dialog_show_error (GTK_WINDOW (window), error, _("Failed to execute command \"%s\"."), command);
+      g_error_free (error);
+    }
+
+  g_string_free (string, TRUE);
+  g_object_unref (G_OBJECT (item));
 
   if (G_LIKELY (xfconf_channel_get_bool (window->channel, "/close-after-execute", FALSE)))
     _xfce_appfinder_window_closed (window);
@@ -780,6 +850,9 @@ _xfce_appfinder_window_load_menu (XfceAppfinderWindow *window,
         return;
       }
 
+  /* connect to the directory changed signal of all menus, force a reload when one of the menus changes */
+  g_signal_connect_swapped (menu, "directory-changed", G_CALLBACK (xfce_appfinder_window_reload), window);
+
   /* Determine menu name */
   name = garcon_menu_element_get_name (GARCON_MENU_ELEMENT (menu));
 
@@ -798,8 +871,6 @@ _xfce_appfinder_window_load_menu (XfceAppfinderWindow *window,
   /* Create category widget */
   if (G_LIKELY (current_counter > 0 && is_category))
     {
-      DBG ("name = %s", name);
-
       if (G_LIKELY (name != NULL))
         {
           button = gtk_radio_button_new_with_label (window->categories_group, name);
@@ -822,73 +893,9 @@ _xfce_appfinder_window_load_menu (XfceAppfinderWindow *window,
 
 
 
-static gpointer
-_xfce_appfinder_window_reload_menu (XfceAppfinderWindow *window)
-{
-  GtkWidget *button;
-  GError    *error = NULL;
-  gint       counter = 0;
-
-  g_return_val_if_fail (XFCE_IS_APPFINDER_WINDOW (window), NULL);
-
-  DBG ("window->menu_filename = %s", window->menu_filename);
-
-  if (G_UNLIKELY (window->menu_filename != NULL))
-    window->menu = garcon_menu_new_for_path (window->menu_filename);
-  else
-    window->menu = garcon_menu_new_applications ();
-
-  if (G_UNLIKELY (window->menu == NULL))
-    {
-      if (G_UNLIKELY (window->menu_filename != NULL))
-        xfce_dialog_show_error (GTK_WINDOW (window), NULL,
-          _("Could not load menu from %s"), window->menu_filename);
-      else
-        xfce_dialog_show_error (GTK_WINDOW (window), NULL,
-          _("Could not load system menu"));
-
-      if (error != NULL)
-        g_error_free (error);
-
-      return NULL;
-    }
-
-  if (!garcon_menu_load (window->menu, NULL, &error))
-    {
-      g_message ("failed to load the menu: %s", error->message);
-      g_error_free (error);
-      return NULL;
-    }
-
-  if (GTK_IS_WIDGET (window->categories_box))
-    gtk_widget_destroy (window->categories_box);
-
-  window->categories_box = gtk_vbox_new (FALSE, 6);
-  gtk_container_add (GTK_CONTAINER (window->categories_alignment), window->categories_box);
-  gtk_widget_show (window->categories_box);
-
-  button = gtk_radio_button_new_with_label (NULL, _("All"));
-  gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
-  g_signal_connect_swapped (button, "toggled", G_CALLBACK (_xfce_appfinder_window_category_changed), window);
-  g_signal_connect (button, "key-press-event", G_CALLBACK (_xfce_appfinder_window_radio_key_pressed), window);
-  gtk_container_add (GTK_CONTAINER (window->categories_box), button);
-  gtk_widget_show (button);
-
-  if (window->current_category == NULL || g_utf8_collate (window->current_category, _("All")) == 0)
-    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
-
-  window->categories_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (button));
-
-  _xfce_appfinder_window_load_menu (window, window->menu, NULL, &counter, TRUE, FALSE);
-
-  return NULL;
-}
-
-
-
 static void
 _xfce_appfinder_window_load_menu_item (XfceAppfinderWindow *window,
-                                       GarconMenuItem        *item,
+                                       GarconMenuItem      *item,
                                        const gchar         *category,
                                        gint                *counter)
 {
@@ -927,9 +934,9 @@ _xfce_appfinder_window_load_menu_item (XfceAppfinderWindow *window,
   command = garcon_menu_item_get_command (item);
 
   if (G_LIKELY (comment != NULL))
-    text = g_strdup_printf ("<b>%s</b>\n%s", name, comment);
+    text = g_markup_printf_escaped ("<b>%s</b>\n%s", name, comment);
   else
-    text = g_strdup_printf ("<b>%s</b>", name);
+    text = g_markup_printf_escaped ("<b>%s</b>", name);
 
   tooltip_str = g_string_new (NULL);
 
@@ -971,6 +978,8 @@ _xfce_appfinder_window_load_menu_item (XfceAppfinderWindow *window,
   if (GDK_IS_PIXBUF (pixbuf))
     g_object_unref (pixbuf);
 
+  g_signal_connect_swapped (item, "changed", G_CALLBACK (xfce_appfinder_window_reload), window);
+
   *counter += 1;
 }
 
@@ -1118,6 +1127,7 @@ _xfce_appfinder_window_visible_func (GtkTreeModel *filter,
   g_free (command);
   g_free (text);
   g_free (category);
+  g_object_unref (G_OBJECT (item));
 
   return visible;
 }
@@ -1140,5 +1150,5 @@ _xfce_appfinder_window_set_category (XfceAppfinderWindow *window,
   gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (window->filter));
   gtk_widget_set_sensitive (window->execute_button, FALSE);
 
-  g_object_notify (G_OBJECT (window), "category");
+  xfconf_channel_set_string (window->channel, "/category", window->current_category);
 }
